/*******************************************
 *  MAIN.CPP - The KINT Interpreter
 *   Copyright 2002-2008 Everett Kaser
 *       All rights reserved.
 *
 *  Do not copy or redistribute without the
 *  author's explicit written permission.
 ********************************************/

#include "main.h"
#include "wkeys.h"

#define TEXT(s) 	(s)
#define	Widen(s)	((char *)s)
#define	UnWiden(s, pw)	strcpy(s, pw)
#define	kscpy(d, s)	strcpy(d, s)
#define	kscat(d, s)	strcat(d, s)
#define	kslen(s)	strlen(s)
#define	kscmp(d, s)	strcmp(d, s)
#define MAXMSGS         1024	// max messages queued at a time in KINTmsgs

extern PBMB   kBuf, pbFirst;

/*******************  Global Variables ********************/

char          OSctrl_cmdSTR[32];
TCHAR         AppDataPath[PATHLEN];	// where to store game user-modified files
KE            mainKE;		// struct for main .KE program
SDL_TimerID   TimerID;
DWORD         IdleCount;
DWORD        *KINTmsgs;
int           KINTmsgadd, KINTmsgrem;
int           PhysicalScreenW, PhysicalScreenH, PhysicalScreenSizeIndex;
TCHAR         PgmPath[PATHLEN], TmpPath[PATHLEN];
TCHAR         PartPath[PATHLEN];	// used ONLY by GetFileName()!!!
TCHAR         AppDataPath[PATHLEN];	// where to store game user-modified files
WORD          MusicPlaying = FALSE;
BOOL          ProgramReady = FALSE;
char          KEname[128];

#define MAX_WAV 10
Mix_Chunk    *Waves[MAX_WAV];
char         *WavName[MAX_WAV];
int           WavHit[MAX_WAV];

SDL_Surface  *screen = NULL;

#ifdef DEBUG
int
printe( char *fmt, ... )
{
  int           ret;
  va_list       ap = NULL;
  char          line[PATHLEN];

#if 0
  if( strncmp( "TODO", fmt, 4 ) != 0 )
    return 0;
#endif
  va_start( ap, fmt );
  ret = vsnprintf( line, PATHLEN, fmt, ap );
  va_end( ap );
  puts( line );
  return ret;
}
#endif

int
stat_all_case( char *path, struct stat *st )
{
  char         *p, *sl, *dot;

  if( ( sl = strrchr( path, '/' ) ) == NULL )	// last slash
    sl = path;
  dot = strrchr( sl, '.' );	// last dot (extension)
  if( stat( path, st ) == 0 )	// normal file name
    return 1;
  if( dot )
  {
    p = dot;
    while( *p++ )
      *p = toupper( *p );
    if( stat( path, st ) == 0 )	// upper extension
      return 1;
  }
  p = sl;
  while( *p++ )
    *p = tolower( *p );
  if( stat( path, st ) == 0 )	// all lower case
    return 1;
  if( dot )
  {
    p = dot;
    while( *p++ )
      *p = toupper( *p );
    if( stat( path, st ) == 0 )	// upper extension
      return 1;
  }
  p = sl;
  while( *p++ )
    *p = toupper( *p );
  if( stat( path, st ) == 0 )	// all upper case
    return 1;
  if( dot )
  {
    p = dot;
    while( *p++ )
      *p = tolower( *p );
    if( stat( path, st ) == 0 )	// lower extension
      return 1;
  }
  return 0;
}

//****************************************************************
//
void
KGetOSversion( BYTE * pMem )
{
  struct utsname un;
  if( uname( &un ) == -1 )
    return;
  sprintf( pMem, "OS %s, Node %s, Release %s, Version %s, Machine %s",
      un.sysname, un.nodename, un.release, un.version, un.machine );
}

//*****************************************************
void
KGetTime( KTIME * pT )
{
  time_t        seconds;
  struct tm    *tm;

  time( &seconds );
  tm = localtime( &seconds );
  pT->y = ( WORD ) tm->tm_year + 1900;
  pT->mo = ( WORD ) tm->tm_mon + 1;
  pT->d = ( WORD ) tm->tm_mday;
  pT->h = ( WORD ) tm->tm_hour;
  pT->mi = ( WORD ) tm->tm_min;
  pT->s = ( WORD ) tm->tm_sec;
}

//******************************************************
void
KTimeToString( BYTE * pMem )
{
  time_t        seconds;
  struct tm    *tm;

  time( &seconds );
  tm = localtime( &seconds );
  sprintf( ( char * ) pMem, "%d-%d-%d %2.2d:%2.2d:%2.2d",
      tm->tm_mon + 1, tm->tm_mday, tm->tm_year + 1900,
      tm->tm_hour, tm->tm_min, tm->tm_sec );
}

//****************************************************************************
// Return a millisecond count which doesn't necessarily bear
// any resemblance to "real" time, but rather is a counter that is
// a reasonably accurate "millisecond counter" that may eventually
// roll over to 0, but which should not do so more often than once a day,
// and who's count value continues upwards (until such time as a roll-over
// might occur).  This is accomplished on the MS-Windows platform using
// the GetTickCount() function.  Other platforms may choose to implement
// this in whatever way seems easiest, while meeting the above criteria.
// This counter value is only used for timing the speed of animations
// and brief delays in actions.  So, reasonable accuracy down to 5 or 10
// milliseconds would be very nice, and the roll-over (if and when it
// occurs) should only cause (at most) a slight burp in any animation
// or screen action that is going on at the moment (but from which it
// should recover almost immediately).
// NOTE: It is IMPORTANT that his count be relative accurate, as it is
// used for generating the elapsed time count for solving puzzles!
DWORD
KGetTicks(  )
{
  DWORD         ticks;
  struct timespec tp;

  clock_gettime( CLOCK_REALTIME, &tp );
  ticks = tp.tv_sec * 1000;
  ticks += tp.tv_nsec / 1000000;
  // printf( "GetTicks %u\n", ticks );
  return ticks;
}

// yield to OS for 'tm' milliseconds
void
KSleep( DWORD tm )
{
  usleep( tm * 1000 );
}

Uint32
timer_callback( Uint32 interval, void *param )
{
  SDL_Event     event;
  SDL_UserEvent userevent;

  userevent.type = SDL_USEREVENT;
  userevent.code = 0;
  userevent.data1 = NULL;
  userevent.data2 = NULL;

  event.type = SDL_USEREVENT;
  event.user = userevent;
  SDL_PushEvent( &event );
  return interval;
}

//*********************************************************
void
StartTimer( DWORD timval )
{
  printe( "StartTimer" );
  // create periodic timer
  if( TimerID == 0 )
    TimerID = SDL_AddTimer( timval, timer_callback, NULL );
}

//*********************************************************
void
StopTimer(  )
{
  printe( "StopTimer" );
  if( TimerID != 0 )
    SDL_RemoveTimer( TimerID );
  TimerID = 0;
}

//*********************************************************
void
StopMusic(  )			// TODO
{
  printe( "TODO - StopMusic" );
}

//****************************************************************************
void
PlayMusic( char *pMem )		// TODO
{
  StopMusic(  );		// stop any previously playing music
  printe( "TODO - PlayMusic" );
}

//***********************************************************
// Return 0 if fail, non-0 if succeed (and set pw, ph, Xppi, Yppi)
// Get everything setup for printing which is to follow
DWORD
KPrintStart( DWORD * pw, DWORD * ph, DWORD * Xppi, DWORD * Yppi )	// TODO
{
  printe( "TODO - KPrintStart" );
  return 0;
}

//*******************************************************
// Shutdown and clean up after printing, we're all through
void
KPrintStop(  )			// TODO
{
  printe( "TODO - KPrintStop" );
}

//*****************************************************
//      return non-0 if successful, 0 if fail
BYTE
KPrintBeginPage(  )		// TODO
{
  printe( "TODO - KPrintBeginPage" );
  return 0;
}

//*****************************************************
//      return non-0 if successful, 0 if fail
BYTE
KPrintEndPage(  )		// TODO
{
  printe( "TODO - KPrintEndPage" );
  return 0;
}

//******************************************************
void
KPrintText( int x, int y, char *str, DWORD clr )	// TODO
{
  printe( "TODO - KPrintText" );
}

//***********************************************************
void
KPrintGetTextSize( char *str, DWORD * dw, DWORD * dh, DWORD * dasc )	// TODO
{
  printe( "TODO - KPrintGetTextSize" );
}

//**************************************************************************
void
KPrintBMB( int xd, int yd, int wd, int hd, PBMB bmS, int xs, int ys, int ws, int hs )	// TODO
{
  printe( "TODO - KPrintBMB" );
}

//**************************************************************************
void
KPrintRect( int xd, int yd, int wd, int hd, DWORD fillcolor, DWORD edgecolor )	// TODO
{
  printe( "TODO - KPrintRect" );
}

//************************************************************************************
void
KPrintFont( WORD hlim, WORD flags, WORD * h, WORD * asc )	// TODO
{
  printe( "TODO - KPrintFont" );
}

//*********************************************************
void
KPrintLines( DWORD * pMem, WORD numpoints, DWORD clr )	// TODO
{
  printe( "TODO - KPrintLines" );
}

//**********************************************************************
void
FlushScreen(  )
{
  printe( "FlushScreen" );
  SDL_UpdateRect( screen, 0, 0, 0, 0 );
}

//****************************************************************
void
KCaptureMouse( BOOL cap )
{
}

//********************************************************************
// Display the NULL-terminated message in 'pMem' using OS resources.
void
SystemMessage( char *pMem )
{
  fprintf( stderr, "SystemMessage: %s\n", pMem );
}

//*********************************************************
// don't allow a program to access anything above the game directory,
// so don't allow absolute paths or .. or // or such.
char         *
GetFileName( char *fname, char *xname )
{
  char         *p;

  if( fname[0] == '/' )
    return NULL;		// fail miserably if they try absolute path
  strcpy( PartPath, fname );
  for( p = PartPath; *p; p++ )
  {
    if( ( *p == '.' && *( p + 1 ) == '.' )
	|| ( *p == '/' && *( p + 1 ) == '/' ) || ( *p == ':' ) )
      return NULL;		// fail miserably if they try some tricky directory 
    if( *p == '\\' )
      *p = '/';			// DOS/Windows path conversion
  }
  strcat( PartPath, xname );	// stick desired file extension on end
  return PartPath;
}

//*********************************************************
//      In the Windows implementation, UI, IMAGES, and BACKGNDS files
//      are all .BMP files, and get that file extendsion added to them
//      automatically.  MUSIC files are assumed to be MIDI files and get .MID.
//      SOUNDS are assumed to be .WAV files, and get .WAV appended.
//      When porting this to another platform (or even under Windows),
//      there's absolutely NO reason why you can't use other graphics file
//      formats (so long as they're 'lossless', so as to preserve the exact
//      colors of the pixels, so that transparency works right), or other
//      music (like MP3) or sound file formats.  As long as you modify
//      the KINT code that opens and loads/plays the files, you can use
//      whatever file formats you want.  The KASM code doesn't care.
//
//      Keep all extensions in this code lowercase !!!
//
// This routine MUST build the full file path in 'TmpPath', as that's
// where the calling functions expect to find it.
void
BuildPath( char *pMem, BYTE folder, BOOL AppData )
{
  char         *p;

  if( AppData )
    kscpy( TmpPath, AppDataPath );
  else
    kscpy( TmpPath, PgmPath );
  switch ( folder & 0x0F )
  {
    case DIR_GAME:		// GAME (home) folder
      kscat( TmpPath, Widen( GetFileName( pMem, "" ) ) );
      break;
    case DIR_UI:
      kscat( TmpPath, TEXT( "UI/" ) );
      if( pMem == NULL )
	break;
      kscat( TmpPath, Widen( pMem ) );
      if( !( folder & DIR_NOEXT ) )
      {
	if( folder & DIR_LOSSY )
	  kscat( TmpPath, TEXT( ".jpg" ) );
	else
	  kscat( TmpPath, TEXT( ".bmp" ) );
      }
      break;
    case DIR_IMAGES:
      kscat( TmpPath, TEXT( "IMAGES/" ) );
      if( pMem == NULL )
	break;
      kscat( TmpPath, Widen( pMem ) );
      if( !( folder & DIR_NOEXT ) )
      {
	if( folder & DIR_LOSSY )
	  kscat( TmpPath, TEXT( ".jpg" ) );
	else
	  kscat( TmpPath, TEXT( ".bmp" ) );
      }
      break;
    case DIR_BACKGNDS:
      kscat( TmpPath, TEXT( "BACKGNDS/" ) );
      if( pMem == NULL )
	break;
      kscat( TmpPath, Widen( pMem ) );
      if( !( folder & DIR_NOEXT ) )
      {
	if( folder & DIR_LOSSY )
	  kscat( TmpPath, TEXT( ".jpg" ) );
	else
	  kscat( TmpPath, TEXT( ".bmp" ) );
      }
      break;
    case DIR_MUSIC:
      // note: in this implementation, all music files are in a "common" folder up one level from the 'home' folder
      // so that other games from Everett Kaser Software can share the same MIDI files.
      kscat( TmpPath, TEXT( "../MUSIC/" ) );
      if( pMem == NULL )
	break;
      kscat( TmpPath, Widen( pMem ) );
      if( !( folder & DIR_NOEXT ) )
	kscat( TmpPath, TEXT( ".mid" ) );
      break;
    case DIR_SOUNDS:
      kscat( TmpPath, TEXT( "WAV/" ) );
      if( pMem == NULL )
	break;
      kscat( TmpPath, Widen( pMem ) );
      if( !( folder & DIR_NOEXT ) )
	kscat( TmpPath, TEXT( ".wav" ) );
      break;
    case DIR_GAME_IMG:
      // note: this is a GRAPHICS file, located in the GAME folder.
      kscat( TmpPath, Widen( GetFileName( pMem, "" ) ) );
      if( !( folder & DIR_NOEXT ) )
      {
	if( folder & DIR_LOSSY )
	  kscat( TmpPath, TEXT( ".jpg" ) );
	else
	  kscat( TmpPath, TEXT( ".bmp" ) );
      }
      break;
    case DIR_UI_COMMON:
      // note: in this implementation, all UI_COMMON files are in a "common" folder up 
      // one level from the 'home' folder so that other games from Everett Kaser Software 
      // can share the same UI files.  This folder contains .UI files and lossyBM (.JPG) files
      kscat( TmpPath, TEXT( "../COMMONUI/" ) );
      if( pMem == NULL )
	break;
      kscat( TmpPath, Widen( pMem ) );
      if( !( folder & DIR_NOEXT ) )
      {
	if( folder & DIR_LOSSY )
	  kscat( TmpPath, TEXT( ".jpg" ) );
	else
	  kscat( TmpPath, TEXT( ".bmp" ) );
      }
      break;
    default:
      break;
  }
  p = TmpPath;
  while( *p )
  {
    if( *p == '\\' )
      *p = '/';
    p++;
  }
}

//*******************************************************************
void
KPlaySound( char *pMem )
{
  int           i, n;

  for( i = 0; i < MAX_WAV; i++ )
    if( Waves[i] != NULL && !strcmp( WavName[i], pMem ) )
    {
      n = i;
      break;
    }
  if( i == MAX_WAV )		// sound was not cached
  {
    for( i = 0; i < MAX_WAV; i++ )
      if( Waves[i] == NULL )
	break;
    if( i == MAX_WAV )		// sound cache is full // TODO
    {
      ;
    }
    else
      n = i;
    if( Kstat( pMem, DIR_SOUNDS, NULL, NULL ) == -1 )	// try various char cases
      return;
    if( ( Waves[n] = Mix_LoadWAV( TmpPath ) ) == NULL )
      return;			// error loading WAV file
    if( WavName[n] )
      free( WavName[n] );
    WavName[n] = strdup( pMem );
    WavHit[n] = 1;		// first hit
  }
  else				// sound is cached
  {
    WavHit[n]++;
  }
  Mix_PlayChannel( -1, Waves[n], 0 );
}

//*******************************************************************
void
KPlaySoundMem( char *pMem, BOOL nostop )	// TODO
{
///  sndPlaySound( pMem, nostop )
  printe( "TODO - KPlaySoundMem" );
}

//*********************************************************************
// store the appropriate End-Of-Line sequence into buffer 'pMem' and
// return the number of bytes written to 'pMem'.
int
GetSystemEOL( BYTE * pMem )
{
  *pMem++ = '\n';
  return 1;
}

//*********************************************************************************
// make a sub-folder of the AppData folder or a sub-folder thereof
// return non-zero if successful, else return zero if fail
int
KMakeFolder( char *fname )
{
  if( !kscmp( fname, TEXT( "../COMMONUI" ) ) )
  {
    kscpy( TmpPath, AppDataPath );
    kscat( TmpPath, TEXT( "../COMMONUI" ) );
  }
  else
  {
    BuildPath( fname, DIR_GAME, TRUE );
  }
  return !mkdir( TmpPath, 0755 );
}

//*********************************************************************************
// Delete an empty folder (can NOT contain ANY files)
void
KDelFolder( char *fname )
{
  struct stat st;

  BuildPath( fname, DIR_GAME, TRUE );
  if( stat_all_case( TmpPath, &st ) )
    rmdir( TmpPath );
}

struct dirent **namelist;
int           namenr;
BOOL          stripEXT;
char         *f_pattern;

int
filter_func( const struct dirent *entry )
{
  // might need FNM_CASEFOLD which is a GNU ext
  return !fnmatch( f_pattern, entry->d_name, 0 );
}

//****************************************************
// Terminate the Kfindfirst()/Kfindnext() process
void
Kfindclose(  )
{
  printe( "Kfindclose" );
  if( namelist != NULL )
  {
    free( namelist );
    namelist = NULL;
  }
}

//******************************************************************
// Search for a filename (possibly containing '*' as wild card)
DWORD
Kfindfirst( char *fname, BYTE * fbuf, BYTE * attrib, BYTE dir )
{
  printe( "Kfindfirst: %s", fname );
  f_pattern = NULL;
  stripEXT = dir != DIR_GAME && !( dir & DIR_NOEXT );
  // if there's an AppData directory tree, search it first
  BuildPath( fname, dir, TRUE );
  if( ( f_pattern = strrchr( TmpPath, '/' ) ) != NULL )
    *f_pattern++ = '\0';
  namenr = scandir( TmpPath, &namelist, filter_func, NULL );
  if( namenr <= 0 )
  {
    BuildPath( fname, dir, FALSE );	// now search the PgmPath directory tree
    if( ( f_pattern = strrchr( TmpPath, '/' ) ) != NULL )
      *f_pattern++ = '\0';
    namenr = scandir( TmpPath, &namelist, filter_func, NULL );
  }
  if( namenr <= 0 )
    return 0xFFFFFFFF;
  namenr--;
  *attrib = ( namelist[namenr]->d_type & DT_DIR ) ? 0x10 : 0;
  strcpy( fbuf, namelist[namenr]->d_name );
  if( stripEXT )
  {
    char         *p = strrchr( fbuf, '.' );
    if( p )
      *p = '\0';
  }
  free( namelist[namenr] );
  return 0;
}

//********************************************************************
// Search for next filename in sequence started by Kfindfirst()
DWORD
Kfindnext( BYTE * fbuf, BYTE * attrib )
{
  printe( "Kfindnext" );
  if( namenr )
  {
    namenr--;
    *attrib = ( namelist[namenr]->d_type & DT_DIR ) ? 0x10 : 0;
    strcpy( fbuf, namelist[namenr]->d_name );
    if( stripEXT )
    {
      char         *p = strrchr( fbuf, '.' );
      if( p )
	*p = '\0';
    }
    free( namelist[namenr] );
    return 0;
  }
  return 0xFFFFFFFF;
}

//*************************************************
void
Kdelete( char *fname, BYTE dir )
{
  struct stat st;

  BuildPath( fname, dir, TRUE );
  if( stat_all_case( TmpPath, &st ) )
    unlink( TmpPath );
}

//*****************************************************
// Only called from SysMakeShareCopy in PMACH.C and Kopen() in MAIN.C
// so always copies from PgmPath
WORD
KCopyFile( char *src, char *dst )
{
  return FALSE;
}

//**********************************************************
// curtype = 0 for normal, 1 for wait, 2 for hotlink
void
KSetCursor( WORD curtype )
{
  printe( "TODO - KSetCursor %d", curtype );
  // must create various shapes first and then call
  // SDL_Cursor *SDL_CreateCursor(Uint8 *data, Uint8 *mask, int w, int h, int hot_x, int hot_y);
}

//**********************************************************
void
KSetTitle( BYTE * pMem )	// Set the window caption
{
  printe( "KSetTitle - %s", ( char * ) pMem );
  SDL_WM_SetCaption( ( char * ) pMem, NULL );
}

DWORD
open_file( int oflags )
{
  struct stat st;

  if( oflags & 0x300 )		// create file
    return creat( TmpPath, 0644 );
  if( oflags & 2 )		// open for read/write
    return open( TmpPath, O_RDWR );
  if( !stat_all_case( TmpPath, &st ) ) // file exists?
    return -1;
  return open( TmpPath, O_RDONLY ); // open readonly
}

//*********************************************************************************
// return file handle if successful, else -1
DWORD
Kopen( char *fname, int oflags, BYTE dir )
{
  int           ret;

  BuildPath( fname, dir, TRUE );	// try AppData first
  if( ( ret = open_file( oflags ) ) != -1 )
  {
    printe( "open %s %x", TmpPath, oflags );
    return ret;
  }
  BuildPath( fname, dir, FALSE );	// try PgmPath second
  if( ( ret = open_file( oflags ) ) != -1 )
  {
    printe( "open %s %x", TmpPath, oflags );
    return ret;
  }
  printe( "open failed for %s %x", TmpPath, oflags );
  return ret;
}

//*********************************************************************************
// return file len if successful, else -1
DWORD
Kstat( char *fname, BYTE dir, DWORD * modtim, BYTE * filtyp )
{
  struct stat   fstats;

  while( 1 )
  {
    BuildPath( fname, dir, TRUE );	// try AppData first
    if( stat_all_case( TmpPath, &fstats ) )
      break;
    BuildPath( fname, dir, FALSE );	// try PgmPath second
    if( stat_all_case( TmpPath, &fstats ) )
      break;
    return 0xFFFFFFFF;
  }
  if( modtim != NULL )
    *modtim = fstats.st_mtime;
  if( filtyp != NULL )
  {
    if( ( fstats.st_mode & S_IFMT ) == S_IFDIR )
      *filtyp = 2;
    else
      *filtyp = 1;
  }
  return fstats.st_size;
}

//********************************************************************************
// In KINT/KASM, 'origin' has these values (in case your language is different):
//              SEEK_SET    0
//              SEEK_CUR    1
//              SEEK_END    2
// return new file pointer position or -1 if error
DWORD
Kseek( DWORD fhandle, DWORD foffset, int origin )
{
  return lseek( ( int ) fhandle, ( long ) foffset, origin );
}

//*********************************************************************************
void
Kclose( DWORD fhandle )
{
  close( ( int ) fhandle );
}

//********************************************************************************
// return 1 if at EOF, 0 if not at EOF, -1 if error
DWORD
Keof( DWORD fhandle )
{
  off_t         pos_current, pos_end;

  pos_current = lseek( fhandle, ( off_t ) 0, SEEK_CUR );
  pos_end = lseek( fhandle, ( off_t ) 0, SEEK_END );
  lseek( fhandle, pos_current, SEEK_SET );
  return ( pos_current == pos_end );
}

//******************************************************
DWORD
Kread( DWORD fhandle, BYTE * pMem, DWORD numbytes )
{
  return read( fhandle, pMem, numbytes );
}

//*******************************************************
DWORD
Kwrite( DWORD fhandle, BYTE * pMem, DWORD numbytes )
{
  return write( fhandle, pMem, numbytes );
}

PBMB
NodeSurface( SDL_Surface * sf )
{
  PBMB          pX;

  if( !sf )
    return NULL;
  pX = ( PBMB ) malloc( sizeof( BMB ) );
  pX->sf = sf;
  pX->pNext = ( void * ) pbFirst;
  if( pbFirst != NULL )
    pbFirst->pPrev = ( void * ) pX;
  pX->pPrev = NULL;
  pbFirst = pX;
  pX->orient = BMO_NORMAL;
  printe( "surface %p l%d w%d h%d", sf->pixels, sf->pitch, sf->w, sf->h );
  return pX;
}

//***************************************
PBMB
CreateBMB( WORD w, WORD h, WORD sysBM )
{
  SDL_Surface  *sf;

  printe( "CreateBMB %dx%d (%d)", w, h, sysBM );
  if( sysBM )
  {
    // create main window
    sf = SDL_SetVideoMode( w, h, 32, SDL_RESIZABLE | SDL_SWSURFACE );
    if( sf )
      screen = sf;
  }
  else
  {
    // printf("-> New Surface %d,%d\n", w, h );
    // create a normal surface
    sf = SDL_CreateRGBSurface( SDL_SWSURFACE, w, h, 32,
	0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000 );
  }
  return NodeSurface( sf );
}

//***************************************
void
DeleteBMB( PBMB pX )
{
  printe( "DeleteBMB" );
  if( pX != NULL )
  {
    if( pX->sf != NULL )
      SDL_FreeSurface( pX->sf );
    if( pX == pbFirst )
    {
      pbFirst = ( PBMB ) pX->pNext;
      if( pbFirst != NULL )
	pbFirst->pPrev = NULL;
    }
    else
    {
      if( pX->pNext != NULL )
	( ( PBMB ) pX->pNext )->pPrev = pX->pPrev;
      if( pX->pPrev != NULL )
	( ( PBMB ) pX->pPrev )->pNext = pX->pNext;
    }
    free( pX );
  }
}

//**********************************************************************
void
InvalidateScreen( int x, int y, int w, int h )
{
  printe( "InvalidateScreen %d,%d %dx%d", x, y, w, h );
  SDL_UpdateRect( screen, x, y, w, h );
}

//**************************************************************************
void
KINT_GetWindowPos( DWORD * r11, DWORD * r12, DWORD * r13, DWORD * r14,
    DWORD * r15 )
{
  printe( "KINT_GetWindowPos" );
  *r15 = 64;			// Position.left
  *r14 = 64;			// Position.top
  *r13 = screen->w;
  *r12 = screen->h;
  *r11 = 1;			// SW_SHOWNORMAL
}

//***********************************************************************
void
KINT_SetWindowPos( DWORD r11, DWORD r12, DWORD r13, DWORD r14, DWORD r15 )
{
  SDL_Surface  *sf;
  /* SW_HIDE = 0 SW_SHOWNORMAL = 1 SW_SHOWMINIMIZED = 2 SW_SHOWMAXIMIZED = 3 SW_SHOWNOACTIVATE = 4 SW_SHOW = 5 SW_MINIMIZE = 6 SW_SHOWMINNOACTIVE = 7 SW_SHOWNA 
   * = 8 SW_RESTORE = 9 SW_SHOWDEFAULT = 10 */
  printe( "KINT_SetWindowPos cmd %ld h %ld w %ld t %ld l %ld",
      r11, r12, r13, r14, r15 );
  if( r12 != 0 && r13 != 0 )
  {
    // resize main window
    sf = SDL_SetVideoMode( r13, r12, 32, SDL_SWSURFACE | SDL_RESIZABLE );
    if( sf )
      screen = sf;
  }
}

//********************************************************************
void
AddMsg( WORD cmd, DWORD arg1, DWORD arg2, DWORD arg3 )
{

  if( KINTmsgs != NULL )
  {
    KINTmsgs[KINTmsgadd * 4] = ( DWORD ) cmd;
    KINTmsgs[KINTmsgadd * 4 + 1] = arg1;
    KINTmsgs[KINTmsgadd * 4 + 2] = arg2;
    KINTmsgs[KINTmsgadd * 4 + 3] = arg3;
    ++KINTmsgadd;
    if( KINTmsgadd >= MAXMSGS )
      KINTmsgadd = 0;
    if( KINTmsgadd == KINTmsgrem )
    {
      --KINTmsgadd;
      if( KINTmsgadd < 0 )
	KINTmsgadd = MAXMSGS - 1;
    }
  }
  else
    KEmsg( &mainKE, cmd, arg1, arg2, arg3 );
}

//********************************************************************
void
SendNextMsg(  )
{
  if( KINTmsgs != NULL && KINTmsgrem != KINTmsgadd )
  {
    KEmsg( &mainKE, ( WORD ) KINTmsgs[KINTmsgrem * 4],
	KINTmsgs[KINTmsgrem * 4 + 1], KINTmsgs[KINTmsgrem * 4 + 2],
	KINTmsgs[KINTmsgrem * 4 + 3] );
    ++KINTmsgrem;
    if( KINTmsgrem >= MAXMSGS )
      KINTmsgrem = 0;
  }
}

//****************************************
void
InitScreen( long minW, long minH )
{

  int           redraw;
  const SDL_VideoInfo *info;
  WORD          ow, oh;

  if( minW == -1 )
  {
    ow = kBuf->sf->w;
    oh = kBuf->sf->h;
    minW = minH = 0;
  }
  else
  {
    ow = ( WORD ) minW;
    oh = ( WORD ) minH;
  }
  info = SDL_GetVideoInfo(  );
  PhysicalScreenW = info->current_w;
  PhysicalScreenH = info->current_h;

  if( kBuf != NULL )
  {
    DeleteBMB( kBuf );
    redraw = 1;
  }
  else
    redraw = 0;

  kBuf = CreateBMB( ( WORD ) minW, ( WORD ) minH, TRUE );
  if( NULL == kBuf )
  {
    SystemMessage( "Failed creating screen BITMAP" );
    exit( 0 );
  }
  if( kBuf->sf->w < 800 || kBuf->sf->h < 600 )
    PhysicalScreenSizeIndex = 0;
  else if( kBuf->sf->w < 1024 || kBuf->sf->h < 768 )
    PhysicalScreenSizeIndex = 1;
  else if( kBuf->sf->w < 1152 || kBuf->sf->h < 864 )
    PhysicalScreenSizeIndex = 2;
  else if( kBuf->sf->w < 1280 || kBuf->sf->h < 1024 )
    PhysicalScreenSizeIndex = 3;
  else if( kBuf->sf->w < 1600 || kBuf->sf->h < 1280 )
    PhysicalScreenSizeIndex = 4;
  else
    PhysicalScreenSizeIndex = 5;

  // initialize the ENTIRE screen bitmap to black
  RectBMB( 0, 0, 0, ow, oh, 0, -1 );

  kBuf->sf->w = ow;
  kBuf->sf->h = oh;

  if( redraw )
  {
// call the SCREEN_CHANGE code in the KE module
    AddMsg( KHDR_SCREEN_CHANGE, 0, 0, 0 );
    SendNextMsg(  );
  }
}

void
SDLCleanup( void )
{
  // Free the sound effects
//  Mix_FreeChunk( scratch );

  // Free the music
//  Mix_FreeMusic( music );

  // Quit SDL_mixer
  Mix_CloseAudio(  );

  // Quit SDL
  SDL_Quit(  );
}

//**********************************************************************
// Called from PMACH.CPP when the .KE wants the program to terminate
// Do whatever's necessary to make the program terminate.
void
KillProgram(  )
{
  if( KINTmsgs != NULL )
  {
    while( KINTmsgs != NULL && KINTmsgadd != KINTmsgrem )
      SendNextMsg(  );
    free( KINTmsgs );
    KINTmsgs = ( DWORD * ) ( KINTmsgadd = KINTmsgrem = 0 );
  }
  // the program is being killed, tell the KASM code so it can clean up
  KEmsg( &mainKE, KHDR_SHUT_DOWN, 0, 0, 0 );
  free( mainKE.pKode );
  SysCleanup(  );
  SDLCleanup(  );
  exit( 0 );
}

DWORD
MapFunc( int key )
{
  switch ( key )
  {
    case SDLK_BACKSPACE:
      return VK_BACK;
    case SDLK_TAB:
      return VK_TAB;
    case SDLK_CLEAR:
      return VK_CLEAR;
    case SDLK_RETURN:
    case SDLK_KP_ENTER:
      return VK_RETURN;
    case SDLK_PAUSE:
      return VK_PAUSE;
    case SDLK_ESCAPE:
      return VK_ESCAPE;
    case SDLK_DELETE:
      return VK_DELETE;
    case SDLK_KP_PERIOD:
      return VK_DECIMAL;
    case SDLK_KP_DIVIDE:
      return VK_DIVIDE;
    case SDLK_KP_MULTIPLY:
      return VK_MULTIPLY;
    case SDLK_KP_MINUS:
      return VK_SUBTRACT;
    case SDLK_KP_PLUS:
      return VK_ADD;
    case SDLK_KP0:
      return VK_NUMPAD0;
    case SDLK_KP1:
      return VK_NUMPAD1;
    case SDLK_KP2:
      return VK_NUMPAD2;
    case SDLK_KP3:
      return VK_NUMPAD3;
    case SDLK_KP4:
      return VK_NUMPAD4;
    case SDLK_KP5:
      return VK_NUMPAD5;
    case SDLK_KP6:
      return VK_NUMPAD6;
    case SDLK_KP7:
      return VK_NUMPAD7;
    case SDLK_KP8:
      return VK_NUMPAD8;
    case SDLK_KP9:
      return VK_NUMPAD9;
    case SDLK_UP:
      return VK_UP;
    case SDLK_DOWN:
      return VK_DOWN;
    case SDLK_RIGHT:
      return VK_RIGHT;
    case SDLK_LEFT:
      return VK_LEFT;
    case SDLK_INSERT:
      return VK_INSERT;
    case SDLK_HOME:
      return VK_HOME;
    case SDLK_END:
      return VK_END;
    case SDLK_PAGEUP:
      return VK_PRIOR;
    case SDLK_PAGEDOWN:
      return VK_NEXT;
    case SDLK_F1:
      return VK_F1;
    case SDLK_F2:
      return VK_F2;
    case SDLK_F3:
      return VK_F3;
    case SDLK_F4:
      return VK_F4;
    case SDLK_F5:
      return VK_F5;
    case SDLK_F6:
      return VK_F6;
    case SDLK_F7:
      return VK_F7;
    case SDLK_F8:
      return VK_F8;
    case SDLK_F9:
      return VK_F9;
    case SDLK_F10:
      return VK_F10;
    case SDLK_F11:
      return VK_F11;
    case SDLK_F12:
      return VK_F12;
    case SDLK_F13:
      return VK_F13;
    case SDLK_F14:
      return VK_F14;
    case SDLK_F15:
      return VK_F15;
    case SDLK_NUMLOCK:
      return VK_NUMLOCK;
    case SDLK_RSHIFT:
    case SDLK_LSHIFT:
      return VK_SHIFT;
    case SDLK_RCTRL:
    case SDLK_LCTRL:
      return VK_CONTROL;
    case SDLK_RALT:
    case SDLK_LALT:
      return VK_MENU;

    case SDLK_EXCLAIM:
    case SDLK_QUOTEDBL:
    case SDLK_HASH:
    case SDLK_DOLLAR:
    case SDLK_AMPERSAND:
    case SDLK_QUOTE:
    case SDLK_LEFTPAREN:
    case SDLK_RIGHTPAREN:
    case SDLK_ASTERISK:
    case SDLK_PLUS:
    case SDLK_COMMA:
    case SDLK_MINUS:
    case SDLK_PERIOD:
    case SDLK_SLASH:
    case SDLK_COLON:
    case SDLK_SEMICOLON:
    case SDLK_LESS:
    case SDLK_EQUALS:
    case SDLK_GREATER:
    case SDLK_QUESTION:
    case SDLK_AT:
    case SDLK_LEFTBRACKET:
    case SDLK_BACKSLASH:
    case SDLK_RIGHTBRACKET:
    case SDLK_CARET:
    case SDLK_UNDERSCORE:
    case SDLK_BACKQUOTE:
    case SDLK_KP_EQUALS:
    case SDLK_CAPSLOCK:
    case SDLK_SCROLLOCK:
    case SDLK_RMETA:
    case SDLK_LMETA:
    case SDLK_LSUPER:
    case SDLK_RSUPER:
    case SDLK_MODE:
    case SDLK_COMPOSE:
    case SDLK_HELP:
    case SDLK_PRINT:
    case SDLK_SYSREQ:
    case SDLK_BREAK:
    case SDLK_MENU:
      return 0;
  }
#if 0
  return VK_F16;
  return VK_F17;
  return VK_F18;
  return VK_F19;
  return VK_F20;
  return VK_F21;
  return VK_F22;
  return VK_F23;
  return VK_F24;
  return VK_CANCEL;
  return VK_CAPITAL;
  return VK_SPACE;
  return VK_SELECT;
  return VK_EXECUTE;
  return VK_SEPARATOR;
  return VK_SNAPSHOT;
  return VK_HELP;
  return VK_LWIN;
  return VK_RWIN;
  return VK_APPS;
  return VK_SCROLL;
  return VK_RSHIFT;
  return VK_LSHIFT;
  return VK_RCONTROL;
  return VK_LCONTROL;
  return VK_RMENU;
  return VK_LMENU;
#endif
  return 0;
}

int           modifiers = 0;

void
process_next_event( SDL_Event * event )
{
  DWORD         cmd, arg1;
  WORD          key;

  switch ( event->type )
  {
    case SDL_ACTIVEEVENT:
      printe( "SDL_ActiveEvent; focus %d gain %d", event->active.state,
	  event->active.gain );
// SDL_APPMOUSEFOCUS       0x01            /**< The app has mouse coverage */
// SDL_APPINPUTFOCUS       0x02            /**< The app has input focus */
// SDL_APPACTIVE           0x04            /**< The application is active */
      break;
    case SDL_KEYDOWN:
      // a keyboard key has been pressed
      key = event->key.keysym.sym;
      if( key == SDLK_q && event->key.keysym.mod & KMOD_CTRL )
	KillProgram(  );
      if( key == SDLK_LCTRL || key == SDLK_RCTRL )
	modifiers |= MK_CONTROL;
      if( key == SDLK_LSHIFT || key == SDLK_RSHIFT )
	modifiers |= MK_SHIFT;
      if( ( key >= SDLK_0 && key <= SDLK_9 ) || ( key >= SDLK_a
	  && key <= SDLK_z ) )
	AddMsg( KHDR_ALPHAKEY_DOWN, ( DWORD ) key, 1, 0 );
      else
	AddMsg( KHDR_KEY_DOWN, ( DWORD ) MapFunc( key ), 1, 0 );
      if( key >= SDLK_a && key <= SDLK_z
	  && event->key.keysym.mod & KMOD_SHIFT )
	key &= ~0x20;		// upper case
      if( key >= SDLK_SPACE && key <= SDLK_z )
	AddMsg( KHDR_CHAR, ( DWORD ) key, 1, 0 );
      break;
    case SDL_KEYUP:
      // a keyboard key has been released
      key = event->key.keysym.sym;
      if( key == SDLK_LCTRL || key == SDLK_RCTRL )
	modifiers &= ~MK_CONTROL;
      if( key == SDLK_LSHIFT || key == SDLK_RSHIFT )
	modifiers &= ~MK_SHIFT;
      // KHDR_KEY_UP and KHDR_KEY_DOWN *MUST* be sent for at least SHIFT, CTL, ALT,
      // TAB, and ENTER, or menus and dialogs will be broken 
      if( ( key >= SDLK_0 && key <= SDLK_9 ) || ( key >= SDLK_a
	  && key <= SDLK_z ) )
	AddMsg( KHDR_ALPHAKEY_UP, ( DWORD ) key, 1, 0 );
      else
	AddMsg( KHDR_KEY_UP, ( DWORD ) MapFunc( key ), 1, 0 );
      break;
    case SDL_MOUSEMOTION:
      // the mouse (pointer) has been moved
      AddMsg( KHDR_POINTER_MOVE, ( DWORD ) event->motion.x,
	  ( DWORD ) event->motion.y, modifiers );
      break;
    case SDL_MOUSEBUTTONDOWN:
    case SDL_MOUSEBUTTONUP:
      printe( "SDL_MouseButtonEvent" );
      switch ( event->button.button )
      {
	case SDL_BUTTON_LEFT:	// the LEFT mouse button has been pressed
	  arg1 = VK_LBUTTON;
	  if( event->type == SDL_MOUSEBUTTONDOWN )
	    modifiers |= MK_LBUTTON;
	  else
	    modifiers &= ~MK_LBUTTON;
	  break;
	case SDL_BUTTON_MIDDLE:	// the MIDDLE mouse button has been pressed
	  arg1 = VK_MBUTTON;
	  if( event->type == SDL_MOUSEBUTTONDOWN )
	    modifiers |= MK_MBUTTON;
	  else
	    modifiers &= ~MK_MBUTTON;
	  break;
	case SDL_BUTTON_RIGHT:	// the RIGHT mouse button has been pressed
	  arg1 = VK_RBUTTON;
	  if( event->type == SDL_MOUSEBUTTONDOWN )
	    modifiers |= MK_RBUTTON;
	  else
	    modifiers &= ~MK_RBUTTON;
	  break;
	case SDL_BUTTON_WHEELUP:	// wheel fwd
	  arg1 = VK_UP;
	  break;
	case SDL_BUTTON_WHEELDOWN:	// wheel back
	  arg1 = VK_DOWN;
	  break;
      }
      cmd = event->type == SDL_MOUSEBUTTONDOWN ? KHDR_KEY_DOWN : KHDR_KEY_UP;
      AddMsg( cmd, arg1, 0, event->button.y << 16 | event->button.x );
      break;
    case SDL_JOYAXISMOTION:
      printe( "SDL_JoyAxisEvent" );
      break;
    case SDL_JOYBALLMOTION:
      printe( "SDL_JoyBallEvent" );
      break;
    case SDL_JOYHATMOTION:
      printe( "SDL_JoyHatEvent" );
      break;
    case SDL_JOYBUTTONDOWN:
      printe( "SDL_JoyButtonDownEvent" );
      break;
    case SDL_JOYBUTTONUP:
      printe( "SDL_JoyButtonUpEvent" );
      break;
    case SDL_SYSWMEVENT:
      printe( "SDL_SysWMEvent" );
      break;
    case SDL_VIDEORESIZE:
      printe( "SDL_ResizeEvent" );
      InitScreen( event->resize.w, event->resize.h );
      AddMsg( KHDR_WINDOW_SIZE, kBuf->sf->w, kBuf->sf->h, K_SIZED );
      if( ProgramReady )
	SendNextMsg(  );
      break;
    case SDL_VIDEOEXPOSE:
      printe( "SDL_ExposeEvent" );
      break;
    case SDL_USEREVENT:
      // the timer time period has expired
      if( ProgramReady )
	AddMsg( KHDR_TIMER, 0, 0, 0 );
      break;
    case SDL_QUIT:
      KillProgram(  );
      break;
  }
}

#define MAX_EVENTS 32

int
main( int argc, char *argv[] )
{
  SDL_Event     events[MAX_EVENTS];
  char         *p;
  int           flags;

  if( argc <= 1 )
  {
#if KINT_HELP
    strcpy( KEname, "help.ke" );
#else
    strcpy( KEname, "main.ke" );
#endif
    getcwd( PgmPath, PATHLEN );
  }
  else
  {
    // if present, the argument must be an absolute path to the .KE file to run 
    kscpy( PgmPath, argv[1] );
    if( ( p = strrchr( PgmPath, '/' ) ) == NULL )
      return -1;
    *p++ = 0;
    strcpy( KEname, p );
  }
  strcat( PgmPath, "/" );
  strcpy( AppDataPath, getenv( "HOME" ) );
  strcat( AppDataPath, "/.eks/" );
  mkdir( AppDataPath, 0700 );	// make sure the AppDataPath directory exists

  printe( "PgmPath %s", PgmPath );
  printe( "AppDataPath %s", AppDataPath );

  if( ( KINTmsgs = ( DWORD * ) malloc( MAXMSGS * 16 ) ) == NULL )
  {
    SystemMessage( "Failed to allocate MSG buf" );
    exit( FALSE );
  }

  // try to load the .KE file
  if( !LoadKEFile( KEname, &mainKE ) )
    exit( FALSE );

  SysStartup(  );

  // Initialize the SDL library 
  if( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER ) < 0 )
  {
    fprintf( stderr, "Couldn't initialize SDL: %s\n", SDL_GetError(  ) );
    exit( FALSE );
  }

  // Initialize SDL_image 
  flags = IMG_INIT_JPG | IMG_INIT_PNG;
  if( ( IMG_Init( flags ) & flags ) != flags )
  {
    fprintf( stderr, "No support for JPG or PNG\n" );
    exit( FALSE );
  }

  // Initialize SDL_mixer
  if( Mix_OpenAudio( 22050, MIX_DEFAULT_FORMAT, 1, 32768 ) == -1 )
    return FALSE;

  printe( "Startup" );

  InitScreen( 0, 0 );
  // If there was an error in setting up the screen
  if( screen == NULL )
  {
    SysCleanup(  );
    return FALSE;
  }

  // call the BEGIN_STARTUP code in the KE module (tell it we're starting up and
  // about to create the main program window)
  KEmsg( &mainKE, KHDR_BEGIN_STARTUP, 0, 0, 0 );

  printe( "KHDR_BEGIN_STARTUP" );
  // call the END_STARTUP code in the KE module (tell it to finish setting up, and
  // that the main program window has now been created.
  KEmsg( &mainKE, KHDR_END_STARTUP, 0, 0, 0 );

  printe( "KHDR_END_STARTUP" );
  // possibly cause the window to be shown on the display.
  // 'mainShowCmd' may get set by the .KE code during KHDR_END_STARTUP using
  // SysGetWindowPos and SysSetWindowPos calls.
  // / ShowWindow(hWndMain, mainShowCmd);

  // send one IDLE message to the KASM code in case it needs to do anything once
  // right up front after everything else has been initialized.
  KEmsg( &mainKE, KHDR_IDLE, 0, 0, 0 );

  ProgramReady = TRUE;

// this is the main "message pump" loop.  While no messages are waiting for the
// program and the IDLE call to the KASM code returns non-0, it keeps sending
// IDLE messages to the KASM code.  If the KASM code returns 0, then no more
// IDLE messages are sent until some Windows message arrives for the program.
  for( ;; )
  {
    while( KINTmsgs != NULL && KINTmsgadd != KINTmsgrem )
    {
      SendNextMsg(  );
    }
    IdleCount = 0;
    // check to see if we can do idle work
    while( !SDL_PollEvent( NULL ) )
    {
      if( !KEmsg( &mainKE, KHDR_IDLE, IdleCount++, 0, 0 ) )
	break;
    }
    // either we have a message, or OnIdle returned false
    // remove several mouse motions events because they slow down mouse selections
    SDL_PeepEvents( events, MAX_EVENTS, SDL_GETEVENT, SDL_MOUSEMOTIONMASK );
    if( SDL_PollEvent( &events[0] ) )
    {
      process_next_event( &events[0] );
    }
  }

  return TRUE;
}
